! This file is part of FD4: Four-Dimensional Distributed Dynamic Data structures
!  
! Copyright 2010 Matthias Lieber, ZIH, TU Dresden, Germany
! matthias.lieber@tu-dresden.de
!  
! FD4 is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!  
! FD4 is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!  
! You should have received a copy of the GNU General Public License
! along with FD4.  If not, see <http://www.gnu.org/licenses/>.

!--------------------------------------------------------------!
!                                                              !
!                        FD4 MPI context                       !
!                                                              !
!--------------------------------------------------------------!

#include "fd4.h"

!! FD4 MPI context.
!!
!! - defines MPI data types according to kinds_mod
!! - communicator, rank, nprocs
!! - helper variables, handles, requests, etc. for MPI routines
!! - other persistent fields related to communication
!!
!! *TODO*:
!! - check if we still need all this, see also $fd4_domain because of (de)allocation
!! - good MPIs provide MPI_SIZEOF and MPI_TYPE_MATCH_SIZE to built types matching
!!   the all Fortran90 types
!!
!! @see kinds_mod
!! @author Matthias Lieber
module fd4_mpi_mod

  use fd4_globaldef_mod
  use stack_mod
  
  USE_MPI
  
  implicit none
  private
  
  INCLUDE_MPIFH
  
  
  public :: fd4_mpi, stack
  public :: MPI_I4K, MPI_I8K, MPI_I_K, MPI_R4K, MPI_R8K, MPI_R_K, MPI_R_K_size
  

  ! MPI data types according to kinds_mod
  
  integer, parameter :: MPI_I4K = MPI_INTEGER    !! 4 byte integer MPI datatype
  !integer, parameter :: MPI_I8K = MPI_2INTEGER   !! 8 byte integer MPI datatyte, don't know if it works!
  integer, parameter :: MPI_I8K = MPI_INTEGER8   !! 8 byte integer MPI datatyte
  integer, parameter :: MPI_I_K = MPI_INTEGER    !! default integer MPI datatyte
  integer, parameter :: MPI_R4K = MPI_REAL              !! 4 byte real MPI datatype
  integer, parameter :: MPI_R8K = MPI_DOUBLE_PRECISION  !! 8 byte real MPI datatyte
  integer, parameter :: MPI_R_K = MPI_DOUBLE_PRECISION  !! default real MPI datatyte

  integer(i_k), parameter :: MPI_R_K_size = 8 !! size in bytes of a variable of type MPI_R_K
  

  type fd4_mpi
    integer :: comm       !! MPI communicator
    integer :: rank       !! MPI rank
    integer :: nprocs     !! number of MPI processes
    
    integer, pointer :: recvcnt(:) !! for MPI_All2allv
    integer, pointer :: recvdis(:) !! for MPI_All2allv
    integer, pointer :: sendcnt(:) !! for MPI_All2allv
    integer, pointer :: senddis(:) !! for MPI_All2allv
    
    integer(i_k), pointer :: sendint(:)  !! to exchange integers with neighbors
    integer(i_k), pointer :: recvint(:)  !! to exchange integers with neighbors
    
    real(r4k), pointer :: sendreal(:)    !! to exchange integers with neighbors
    real(r4k), pointer :: recvreal(:)    !! to exchange integers with neighbors

    type(stack) :: requests               !! requests for non-blocking communication

    integer, pointer :: statuses(:,:)     !! array of statuses
    
  end type fd4_mpi

end module fd4_mpi_mod
